@@ -1,27 +1,14 @@
-/********************************************************************
-Copyright 2014  Martin Gräßlin <mgraesslin@kde.org>
+/*
+    SPDX-FileCopyrightText: 2014 Martin Gräßlin <mgraesslin@kde.org>
 
-This library is free software; you can redistribute it and/or
-modify it under the terms of the GNU Lesser General Public
-License as published by the Free Software Foundation; either
-version 2.1 of the License, or (at your option) version 3, or any
-later version accepted by the membership of KDE e.V. (or its
-successor approved by the membership of KDE e.V.), which shall
-act as a proxy defined in Section 6 of version 3 of the license.
-
-This library is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-Lesser General Public License for more details.
-
-You should have received a copy of the GNU Lesser General Public
-License along with this library.  If not, see <http://www.gnu.org/licenses/>.
-*********************************************************************/
+    SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
+*/
 #include "datadevice.h"
-#include "dataoffer.h"
 #include "datasource.h"
 #include "surface.h"
 #include "wayland_pointer_p.h"
+// Qt
+#include <QPointer>
 // Wayland
 #include <wayland-client-protocol.h>
 
@@ -29,8 +16,7 @@ namespace KWayland
 {
 namespace Client
 {
-
-class DataDevice::Private
+class Q_DECL_HIDDEN DataDevice::Private
 {
 public:
     explicit Private(DataDevice *q);
@@ -38,10 +24,17 @@ public:
 
     WaylandPointer<wl_data_device, wl_data_device_release> device;
     QScopedPointer<DataOffer> selectionOffer;
+    struct Drag {
+        QPointer<DataOffer> offer;
+        QPointer<Surface> surface;
+    };
+    Drag drag;
 
 private:
     void dataOffer(wl_data_offer *id);
     void selection(wl_data_offer *id);
+    void dragEnter(quint32 serial, const QPointer<Surface> &surface, const QPointF &relativeToSurface, wl_data_offer *dataOffer);
+    void dragLeft();
     static void dataOfferCallback(void *data, wl_data_device *dataDevice, wl_data_offer *id);
     static void enterCallback(void *data, wl_data_device *dataDevice, uint32_t serial, wl_surface *surface, wl_fixed_t x, wl_fixed_t y, wl_data_offer *id);
     static void leaveCallback(void *data, wl_data_device *dataDevice);
@@ -55,18 +48,12 @@ private:
     DataOffer *lastOffer = nullptr;
 };
 
-const wl_data_device_listener DataDevice::Private::s_listener = {
-    dataOfferCallback,
-    enterCallback,
-    leaveCallback,
-    motionCallback,
-    dropCallback,
-    selectionCallback
-};
+const wl_data_device_listener DataDevice::Private::s_listener =
+    {dataOfferCallback, enterCallback, leaveCallback, motionCallback, dropCallback, selectionCallback};
 
 void DataDevice::Private::dataOfferCallback(void *data, wl_data_device *dataDevice, wl_data_offer *id)
 {
-    auto d = reinterpret_cast<Private*>(data);
+    auto d = reinterpret_cast<Private *>(data);
     Q_ASSERT(d->device == dataDevice);
     d->dataOffer(id);
 }
@@ -78,41 +65,61 @@ void DataDevice::Private::dataOffer(wl_data_offer *id)
     Q_ASSERT(lastOffer->isValid());
 }
 
-void DataDevice::Private::enterCallback(void *data, wl_data_device *dataDevice, uint32_t serial, wl_surface *surface, wl_fixed_t x, wl_fixed_t y, wl_data_offer *id)
+void DataDevice::Private::enterCallback(void *data,
+                                        wl_data_device *dataDevice,
+                                        uint32_t serial,
+                                        wl_surface *surface,
+                                        wl_fixed_t x,
+                                        wl_fixed_t y,
+                                        wl_data_offer *id)
 {
-    Q_UNUSED(data)
-    Q_UNUSED(dataDevice)
-    Q_UNUSED(serial)
-    Q_UNUSED(surface)
-    Q_UNUSED(x)
-    Q_UNUSED(y)
-    Q_UNUSED(id)
+    auto d = reinterpret_cast<Private *>(data);
+    Q_ASSERT(d->device == dataDevice);
+    d->dragEnter(serial, QPointer<Surface>(Surface::get(surface)), QPointF(wl_fixed_to_double(x), wl_fixed_to_double(y)), id);
+}
+
+void DataDevice::Private::dragEnter(quint32 serial, const QPointer<Surface> &surface, const QPointF &relativeToSurface, wl_data_offer *dataOffer)
+{
+    drag.surface = surface;
+    Q_ASSERT(*lastOffer == dataOffer);
+    drag.offer = lastOffer;
+    lastOffer = nullptr;
+    Q_EMIT q->dragEntered(serial, relativeToSurface);
 }
 
 void DataDevice::Private::leaveCallback(void *data, wl_data_device *dataDevice)
 {
-    Q_UNUSED(data)
-    Q_UNUSED(dataDevice)
+    auto d = reinterpret_cast<Private *>(data);
+    Q_ASSERT(d->device == dataDevice);
+    d->dragLeft();
+}
+
+void DataDevice::Private::dragLeft()
+{
+    if (drag.offer) {
+        delete drag.offer;
+    }
+    drag = Drag();
+    Q_EMIT q->dragLeft();
 }
 
 void DataDevice::Private::motionCallback(void *data, wl_data_device *dataDevice, uint32_t time, wl_fixed_t x, wl_fixed_t y)
 {
-    Q_UNUSED(data)
-    Q_UNUSED(dataDevice)
-    Q_UNUSED(time)
-    Q_UNUSED(x)
-    Q_UNUSED(y)
+    auto d = reinterpret_cast<Private *>(data);
+    Q_ASSERT(d->device == dataDevice);
+    Q_EMIT d->q->dragMotion(QPointF(wl_fixed_to_double(x), wl_fixed_to_double(y)), time);
 }
 
 void DataDevice::Private::dropCallback(void *data, wl_data_device *dataDevice)
 {
-    Q_UNUSED(data)
-    Q_UNUSED(dataDevice)
+    auto d = reinterpret_cast<Private *>(data);
+    Q_ASSERT(d->device == dataDevice);
+    Q_EMIT d->q->dropped();
 }
 
 void DataDevice::Private::selectionCallback(void *data, wl_data_device *dataDevice, wl_data_offer *id)
 {
-    auto d = reinterpret_cast<Private*>(data);
+    auto d = reinterpret_cast<Private *>(data);
     Q_ASSERT(d->device == dataDevice);
     d->selection(id);
 }
@@ -121,13 +128,13 @@ void DataDevice::Private::selection(wl_data_offer *id)
 {
     if (!id) {
         selectionOffer.reset();
-        emit q->selectionCleared();
+        Q_EMIT q->selectionCleared();
         return;
     }
     Q_ASSERT(*lastOffer == id);
     selectionOffer.reset(lastOffer);
     lastOffer = nullptr;
-    emit q->selectionOffered(selectionOffer.data());
+    Q_EMIT q->selectionOffered(selectionOffer.data());
 }
 
 DataDevice::Private::Private(DataDevice *q)
@@ -151,6 +158,9 @@ DataDevice::DataDevice(QObject *parent)
 
 DataDevice::~DataDevice()
 {
+    if (d->drag.offer) {
+        delete d->drag.offer;
+    }
     release();
 }
 
@@ -179,7 +189,8 @@ void DataDevice::startDragInternally(quint32 serial, Surface *origin, Surface *i
     startDrag(serial, nullptr, origin, icon);
 }
 
-namespace {
+namespace
+{
 static wl_data_source *dataSource(const DataSource *source)
 {
     if (!source) {
@@ -191,11 +202,7 @@ static wl_data_source *dataSource(const DataSource *source)
 
 void DataDevice::startDrag(quint32 serial, DataSource *source, Surface *origin, Surface *icon)
 {
-    wl_data_device_start_drag(d->device,
-                              dataSource(source),
-                              *origin,
-                              icon ? (wl_surface*)*icon : nullptr,
-                              serial);
+    wl_data_device_start_drag(d->device, dataSource(source), *origin, icon ? (wl_surface *)*icon : nullptr, serial);
 }
 
 void DataDevice::setSelection(quint32 serial, DataSource *source)
@@ -213,12 +220,22 @@ DataOffer *DataDevice::offeredSelection() const
     return d->selectionOffer.data();
 }
 
-DataDevice::operator wl_data_device*()
+QPointer<Surface> DataDevice::dragSurface() const
+{
+    return d->drag.surface;
+}
+
+DataOffer *DataDevice::dragOffer() const
+{
+    return d->drag.offer;
+}
+
+DataDevice::operator wl_data_device *()
 {
     return d->device;
 }
 
-DataDevice::operator wl_data_device*() const
+DataDevice::operator wl_data_device *() const
 {
     return d->device;
 }
